/* isdrush.c - draw ispr file pictures
 * Copyright (C) 2000 Tachibana Eriko
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307
 * USA
 */

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/dos.h>
#include <sys/iocs.h>

#include "isd_header.h"
#include "sdr.h"


/* External functions */
extern int dos_beta (int fd,   int frame_num, int frame_size, char* buf);
extern int sdr_beta (SDR* sdr, int frame_num, int frame_size, char* buf);
extern int dos_yat  (int fd,   int frame_num, int* tbl_ptr,   char* buf);
extern int sdr_yat  (SDR* sdr, int frame_num, int* tbl_ptr,   char* buf);


/* File scope variables */
static const char program[] = "isdrush";
static const char version[] = "1.02";
static const char date[] = "2000/08/15";
static const char author[] = "Ԃq";

static int init_sc_done;
static char* argv0;


/* File scope function */
static int
init_screen (int no_t_off)
{
  int n;

  if (init_sc_done > 0)
    {
      _iocs_b_bpoke ((char *)0x93c, 12);
      _iocs_g_clr_on ();
      _iocs_b_bpoke ((char *)0x93c, 16);
    }
  if (init_sc_done)
    return init_sc_done;		/* ς */

  n = _iocs_tgusemd (0, -1);
  if (n == 1 || n == 2)
    return init_sc_done = -1;		/* OtBbNʎgp */

  /* ʃ[h̐ݒ */
  if (_dos_c_width (-1) != 0)
    _dos_c_width (0);
  _iocs_b_bpoke ((char *)0x93c, 12);	/* 64K 31kHz */
  _iocs_g_clr_on ();
  _iocs_b_bpoke ((char *)0x93c, 16);

  /* \W𒆉ɐݒ */
  {
    int i;
    for (i = 3; i >= 0; i--)
      _iocs_scroll (i, -128, 0);
  }

#ifndef NO_GRAPHIC_MASK
  /* }XN` */
  __asm __volatile ("move.l #'gm'<<16+$ff89,d1\n\t"
		    "moveq #$0e,d0\n\t"
		    "trap #15"		/* gm mask set */
		    : /* no outputs */
		    : /* no outputs */
		    : "d0", "d1");
#endif

  if (no_t_off)
    {
      _iocs_b_clr_st ();
      _iocs_b_down (32);
    }

  return init_sc_done = 1;		/*  */
}


/* File scope function */
static int
isd_rush (const char* filename, int no_t_off, int sdr_flag)
{
  FILE* fp;
  IsdHeader hed;
  int buf_size, yat_flag, frame_num;
  int* tbl_buf = NULL;
  char* buf;
  int rc = 0;

  if (strcmp (filename, "-") == 0)
    {
      filename = "<stdin>";
      fp = stdin;
    }
  else if ((fp = fopen (filename, "rb")) == NULL)
    {
      perror (filename);
      return rc;
    }

  if (fread (&hed, sizeof hed, 1, fp) != 1)
    {
      perror (filename);
      fclose (fp);
      return rc;
    }

  if (memcmp (ISPR_ID, hed.id, sizeof (hed.id)) != 0)
    {
      eprintf ("%s: ISPR f[^ł͂܂.\n", filename);
      fclose (fp);
      return rc;
    }

  if (memcmp ("3.0", hed.version, sizeof (hed.version)) == 0
   || memcmp ("4.0", hed.version, sizeof (hed.version)) == 0)
    {
      yat_flag = 0;			/* k */
    }
  else if (memcmp ("3.5", hed.version, sizeof (hed.version)) == 0
	|| memcmp ("4.5", hed.version, sizeof (hed.version)) == 0)
    {
      yat_flag = 1;			/* YAT k */
    }
  else
    {
      eprintf ("%s: Ή̃o[Wł.\n", filename);
      fclose (fp);
      return rc;
    }

  if (hed.x_length > 512
   || hed.y_length > 512
   || hed.frame_size & 0x3ff
   || (!yat_flag && hed.header_size != sizeof (IsdHeader)))
    {
      eprintf ("%s: Ή̃f[^`ł.\n", filename);
      fclose (fp);
      return rc;
    }

  if (yat_flag)
    {
      int tbl_size, i;

      frame_num = hed.header_size;
      tbl_size = frame_num * sizeof (int);

      if ((tbl_buf = malloc (tbl_size)) == NULL)
	{
	  perror (NULL);
	  fclose (fp);
	  return rc;
	}
      if (fseek (fp, -tbl_size, SEEK_END) < 0
       || fread (tbl_buf, tbl_size, 1, fp) != 1)
	{
	  perror (filename);
	  fclose (fp);
	  return rc;
	}

      buf_size = 0;
      for (i = 0; i < frame_num; ++i)
	buf_size = max (buf_size, tbl_buf[i]);
    }
  else
    {
      if ((frame_num = filelength (fileno (fp))) < 0)
	{
	  perror (filename);
	  fclose (fp);
	  return rc;
	}
      frame_num -= hed.header_size;
      frame_num /= hed.frame_size;
      buf_size = hed.frame_size;
    }

  if ((int) (buf = _dos_malloc2 (2, buf_size)) < 0)
    {
      errno = ENOMEM;
      perror (NULL);
      free (tbl_buf);
      fclose (fp);
      return rc;
    }

  if (_dos_seek (fileno (fp), sizeof (IsdHeader), 0) < 0
   || init_screen (no_t_off) < 0)
    {
      _dos_mfree (buf);
      free (tbl_buf);
      fclose (fp);
      return rc;
    }

  /* Z^O */
  {
    int x = (hed.x_length - 512) / 2 - 128;
    int y = (hed.y_length > 128) ? (hed.y_length - 512) / 2
				 : (hed.y_length - 256);
    int i;

    for (i = 3; i >= 0; i--)
      _iocs_scroll (i, x, y);
  }

  if (!no_t_off)
    _iocs_vc_r2 (_iocs_vc_r2 (-1) & ~0x20);

  {
    SDR* sdr;
    int done = 0;

    if (sdr_flag && (sdr = SDRopen (filename)) != NULL)
      {
	if (sdr_flag == 2)
	  eprintf ("SDRopen(%s) = $%0p\n", filename, sdr);

	if (SDRread (sdr, buf, sizeof (IsdHeader)) == sizeof (IsdHeader))
	  {
	    int ssp = _dos_super (0);

	    rc = yat_flag ? sdr_yat (sdr, frame_num, tbl_buf, buf)
			  : sdr_beta (sdr, frame_num, hed.frame_size, buf);

	    _dos_super (ssp);
	    done = 1;
	  }
	else if (sdr_flag == 2)
	  eprintf ("SDRread() ERROR!!\n");

	SDRclose (sdr);
      }
    if (!done)
      {
	int ssp = _dos_super (0);

	rc = yat_flag ? dos_yat (fileno (fp), frame_num, tbl_buf, buf)
		      : dos_beta (fileno (fp), frame_num, hed.frame_size, buf);

	_dos_super (ssp);
      }
  }

  if (!no_t_off)
    _iocs_vc_r2 (_iocs_vc_r2 (-1) | 0x20);

  _dos_mfree (buf);
  free (tbl_buf);
  fclose (fp);

  return rc;
}


/* File scope function */
static void
print_usage (void)
{
  printf ("%s version %s  %s %s.\n", program, version, date, author);
  printf ("usage: %s [option] [--] [filename] ...\n", argv0);
  printf ("option:\n"
	  "\t-g\tIɃOtBbNʂȂ.\n"
	  "\t-t\tsɃeLXgʂȂ.\n"
	  "\t-s\tSCSI IOCS Œڃt@Cǂݍ(SCSI Direct Read).\n");
}

/* File scope function */
static void
print_version (void)
{
  printf ("%s version %s\n", program, version);
}


/* Function */
int
main (int argc, char* argv[])
{
  int i;
  int count = 0;
  int no_g_off = 0;
  int no_t_off = 0;
  int sdr_flag = 0;
  int no_opt = 0;

  argv0 = argv[0];

  for (i = 1; i < argc; ++i)
    {
      char* p = argv[i];

      if (*p == '\0')
	continue;

      if (*p++ != '-' || *p == '\0' || no_opt)
	{
	  ++count;
	  if (isd_rush (--p, no_t_off, sdr_flag))
	    break;
	}

      else if (p[0] == '-' && p[1] == '\0')
	no_opt = 1;			/* -- */

      else if (p[0] == 's' && p[1] == '\0')
	sdr_flag = 1;			/* -s */

      else if (p[0] == 'S' && p[1] == '\0')
	sdr_flag = 2;			/* -S */

      else if (p[0] == 'g' && p[1] == '\0')
	no_g_off = 1;			/* -g */

      else if (p[0] == 't' && p[1] == '\0')
	no_t_off = 1;			/* -t */

      else if (*p == '?' || *p == 'h' || *p == 'H' || strcmp (p, "-help") == 0)
	{
	  print_usage ();		/* gp@\ */
	  return EXIT_SUCCESS;
	}

      else if (*p == 'v' || *p == 'V' || strcmp (p, "-version") == 0)
	{
	  print_version ();		/* o[W\ */
	  return EXIT_SUCCESS;
	}

      else
	{
	  print_usage ();		/* sȈ */
	  return EXIT_FAILURE;
	}
    }

  if (count == 0)
    {
      if (isatty (STDIN_FILENO))
	{
	  print_usage ();		/* ȗ */
	  return EXIT_FAILURE;
	}
      else
	isd_rush ("-", no_t_off, sdr_flag);
    }

  if (!no_g_off && init_sc_done > 0)
    {
      _iocs_vc_r2 (_iocs_vc_r2 (-1) & ~0x1f);

#ifndef NO_GRAPHIC_MASK
  /* }XN */
  __asm __volatile ("move.l #'gm'<<16+$ff8a,d1\n\t"
		    "moveq #$0e,d0\n\t"
		    "trap #15"		/* gm mask clear */
		    : /* no outputs */
		    : /* no outputs */
		    : "d0", "d1");
#endif
    }

  return EXIT_SUCCESS;
}

/* EOF */
